/*
For information on this software or to report bugs please visit http://www.ranton.org
or email ranton.org

Copyright (c) 2002, Richard Neal Anton
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

*	Redistributions of source code must retain the above copyright notice, this list
	of conditions and the following disclaimer. 

*	Redistributions in binary form must reproduce the above copyright notice, this list
	of conditions and the following disclaimer in the documentation and/or other materials
	provided with the distribution. 

*	The name of the copyright holder may not be used to endorse or promote products 
	derived from this software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>
#include <ctype.h>

void msg_error(char *str, char *filename, int error);

static long ProcessFile(char *filename, void *ptr);
static long sum_file(char *buffer,long len, char *filename);
// this returns the value parsed from the line,
// or zero if no value is parsed out successfully.
static long process_line(char *filename, long lineNum, char *line, long len);

static int verbose = 0;

#ifndef TRUE
#define TRUE 1
#define FALSE 0
#endif

#define kNewline 10

int main(int argc, char **argv)
{
	int count;
	char *filename;
	long total = 0;

	if( argc < 2 )
	{
		fprintf(stderr,"Usage: %s <files>\n",argv[0]);
		exit(1);
	}

	count = 1;

	if( !strcmp(argv[1], "-v") )
	{
		verbose = 1;
		count++;
		if( argc < 3 )
		{
			fprintf(stderr,"Usage: %s <files>\n",argv[0]);
			exit(1);
		}
	}

start:
	{
		filename = argv[count];
		total += ProcessFile( filename,NULL );
		count++;
	}
	if( count < argc )
		goto start;
	

	// print out the total total
	printf("Total: $%ld\n",total);

	return 0;
}

static long ProcessFile(char *filename, void *ptr)
{
	FILE *file = NULL;
	char *buffer = NULL;
	long len;
	struct stat stat_rec;
	long sum;

	if( stat(filename, &stat_rec) )
	{
		msg_error("Error from stat",filename,errno);
		goto error;
	}
	if( S_ISDIR( stat_rec.st_mode ) )
	{
		fprintf(stderr,"%s is a directory, skipping\n",filename);
		goto error;
	}

	if( verbose )
		printf("%s\n",filename);

	file = fopen( filename, "rb" );
	if( !file )
	{
		msg_error( "Error opening file for reading",filename,errno);
		goto error;
	}

	if( fseek(file,0,SEEK_END) )
	{
		msg_error( "Error seeking",filename,errno);
		goto error;
	}

	len = ftell(file);
	if( len == EOF )
	{
		msg_error( "Error getting file position",filename,errno);
		goto error;
	}
	
	if( fseek(file,0,SEEK_SET) )
	{
		msg_error("Error seeking",filename,errno);
		goto error;
	}

	// add one so that we can put a newline on
	// the end if we want to simplify processing.
	buffer = (char *)malloc( sizeof(char) * (len*1) );
	if( !buffer )
	{
		msg_error("Couldn't get buffer",filename,errno);
		goto error;
	}
	
	if( fread( (void *)buffer, len, 1, file ) != 1 )
	{
		msg_error("Error reading",filename,errno);
		goto error;
	}

	if( fclose(file) )
		msg_error( "Error closing file",filename,errno);
	file = NULL;

	// for simplicity of code, put a newline on the end
	// if it doesn't have one.
	if( buffer[len-1] != kNewline )
		buffer[len] = kNewline;
	
	sum = sum_file(buffer,len,filename);

error:
	// free the buffer.
	if( buffer )
		free( buffer );
	buffer = NULL;
	// close the file if needed
	if( file )
		(void)fclose(file);
	file = NULL;


	return sum;
}


// this returns the value parsed from the line,
// or zero if no value is parsed out successfully.
static long process_line(char *filename, long lineNum, char *line, long len)
{
	long value = 0;
	char *ptr = line + len-1;
	int allWhite = TRUE;

	// first off, if it is commented out then ignore it
	if( *line == '#' )
		return 0;
	
	while( ptr >= line )
	{
		if( !isspace(*ptr) )
			allWhite = FALSE;
		if( *ptr == '$' )
			break;
		else
			ptr--;
	}
	// blank line, ignore it.
	if( allWhite )
		return 0;
	
	if( ptr < line )
		goto parse;
	// *ptr is the dollar sign,
	// so skip it, then start
	// calculating the value.
	ptr++;
	while( *ptr != kNewline && !isspace(*ptr) )
	{
		
		if( !isdigit(*ptr) )
			goto parse;
		value *= 10;
		value += *ptr - '0';
		ptr++;
	}
	
	return value;
parse:
	// parse error.
	fprintf(stderr,"Parse error on line %ld of %s\n",lineNum,filename);
	return 0;
}


static long sum_file(char *buffer,long len, char *filename)
{
	char *ptr = buffer, *start;
	long count, lineLen;
	long sum, lineNum;

	
	start = ptr;
	count = 0;
	lineLen = 0;
	sum = 0;
	lineNum = 1;
	while(count < len)
	{
		// if we have hit a newline,
		// search back til we get to a dollar
		// sign or the start of the line.
		if( *ptr == kNewline )
		{
			sum += process_line(filename,lineNum, start,lineLen);
			// reset for the next go.
			start = ptr + 1;
			lineLen = 0;
			lineNum++;
		}
		else
		{
			lineLen++;
		}	
		ptr++;
		count++;
	}
	// if we have anything past the last newline,
	// then deal with it.
	if( lineLen )
		sum += process_line(filename,lineNum,start,lineLen);
	// print the filename and the total
	printf("%s: $%ld\n",filename, sum);


	return sum;
}

void msg_error(char *str, char *filename, int error)
{
	fprintf(stderr,"%s: %s, %s\n",
		str,
		filename,
		strerror(error)
		);
}

